"""ber.py
"""
# Standard library imports

# Third party imports

# Local application imports
from common import hstr as _hstr

def parse(tlv_hstr: str, recursive=True):
    """parse: Parses a TLV string into a (nested) list of (T,L,V) tuples and a remainder string
    """
    remainder = _hstr.clean(tlv_hstr)
    output = []
    while remainder != '':
        # Ignoring 00 bytes before or after TLV objects
        while remainder[0:2] == '00':
            remainder = remainder[2:]

        T, remainder = _decode_tag(remainder)
        if T == '':
            break

        L, remainder = _decode_length(remainder)
        if L == '':
            break

        V, remainder = _decode_value(L, remainder)
        if V == '':
            break

        # Constructed object
        if recursive and _is_tag_constructed(T):
            V, _ = parse(V)

        output.append((T, L, V))

    return (output, remainder)

def parse_to_dict(tlv_hstr: str):
    """
    """
    tlv_object_list, remainder = parse(tlv_hstr, recursive=True)

    return _tlv_object_list_to_dict(tlv_object_list)

def parse_dol(tlv_hstr: str):
    """parse_dol():
    """
    remainder = _hstr.clean(tlv_hstr)
    output = []
    while remainder != '':
        # Ignoring 00 bytes before or after TLV objects
        while remainder[0:2] == '00':
            remainder = remainder[2:]

        T, remainder = _decode_tag(remainder)
        if T == '':
            break

        L, remainder = _decode_length(remainder)
        if L == '':
            break

        output.append((T, L))

    return (output, remainder)

def find(tag, tlv_object):
    """find: finds a specific tag in a list of objects
    """
    tag = _hstr.clean(tag)
    if isinstance(tlv_object, tuple):
        return _find_in_tuple(tag, tlv_object)
    elif isinstance(tlv_object, list):
        return _find_in_list(tag, tlv_object)
    else:
        return None

def encode_length(data: str):
    """encode_length(): encode length of 'data' per BER-TLV encoding rules
    """
    data = _hstr.clean(data, command_name='encode_length')
    nr_nibbles = len(data)
    nr_bytes = nr_nibbles//2

    if nr_nibbles % 2 != 0:
        # Error if input data is not an even number of nibbles (i.e., integer number of bytes)
        return None
    elif nr_bytes <= 127:
        # Definite length, short format
        return F"{nr_bytes:02X}"
    else:
        # Definite length, long format
        l_value = F"{nr_bytes:X}"
        if len(l_value) % 2 != 0:
            # Left pad with 0 to have even number of nibbles
            l_value = F"0{l_value}"
        
        l_header = F"{128+len(l_value)//2:02X}"

        return F"{l_header}{l_value}"

def encode(tag: str, data: str) -> str:
    """encode(): encode tag and data into BER-TLV object
    """
    tag = _hstr.clean(tag)
    data = _hstr.clean(data)

    return F"{tag}{encode_length(data)}{data}"


#
# Helper functions (assume a clean 'hstr' as input)
#

def _find_in_tuple(tag, tlv_object):
    T, _, V = tlv_object
    if T == tag:
        return tlv_object
    elif isinstance(V, list):
        return _find_in_list(tag, V)
    else:
        return None

def _find_in_list(tag, tlv_object_list):
    for tlv_object in tlv_object_list:
        r = _find_in_tuple(tag, tlv_object)
        if r is not None:
            return r
    return None

def _is_tag_constructed(tag: str) -> bool:
    return _hstr.and_(tag[0:2], '20') == '20'

def _decode_tag(data: str) -> tuple[str, str]:
    # Input data must be at least 1 byte (i.e., 2 nibbles)
    if len(data) < 2:
        return ('', data)
    
    # Input data must be byte sequence (i.e., even number of nibbles)
    if len(data) % 2 == 1:
        return ('', data)

    # Determining the length of the tag field.
    if _hstr.and_(data[0:2], '1F') != '1F':
        # The tag field is only one byte.
        return (data[0:2], data[2:])
    else:
        # The tag field has more than one byte.
        index = 2

        # Looking for subsequent bytes (i.e., whether most significant bit of current byte is 1)
        while _hstr.and_(data[index:index+2], '80') == '80':
            index += 2
        
        return (data[0:index+2], data[index+2:])


def _decode_length(data: str) -> tuple[str, str]:
    # Input data must be at least 1 byte (i.e., 2 nibbles)
    if len(data) < 2:
        return ('', data)
    
    # Input data must be byte sequence (i.e., even number of nibbles)
    if len(data) % 2 == 1:
        return ('', data)

    # Determining the length of the length field.
    if _hstr.and_(data[0:2], '80') == '00':
        # The length field is only one byte (definite, short)
        return (data[0:2], data[2:])
    elif data[0:2] in ['80', 'FF']:
        # The length field is only one byte (indefinite or reserved)
        return (data[0:2], data[2:])
    else:
        # The length field has more than one byte (definite, long)
        nr_bytes = int(_hstr.and_(data[0:2], '7F'), 2)
        return (data[0:2+2*nr_bytes], data[2+2*nr_bytes:])


def _decode_value(length: str, data: str) -> tuple[str, str]:
    # Input data must be at least 1 byte (i.e., 2 nibbles)
    if len(data) < 2:
        return ('', data)
    
    # Input data must be byte sequence (i.e., even number of nibbles)
    if len(data) % 2 == 1:
        return ('', data)

    if len(length) == 2:
        if length == '80':
            # Length is of indefinite form i.e., data ends with 2 end-of-contents (0x00) octets
            end_pos = data.find('0000')
            if end_pos == -1:
                # Could not find EOC octets
                return ('', data)
            else:
                return (data[0:end_pos], data[end_pos+4:])
        elif length == 'FF':
            # Length value is 'reserved'
            return ('', data)
        else:
            # Length is definite, short
            nr_bytes = int(length, 16)
            if len(data) < 2 * nr_bytes:
                # The data field is too short
                return ('', data)
            else:
                return (data[0:2*nr_bytes], data[2*nr_bytes:])
    else:
        # Length is definite, long
        nr_bytes = int(length[2:], 16)
        if len(data) < 2 * nr_bytes:
            # The data field is too short
            return ('', data)
        else:
            return (data[0:2*nr_bytes], data[2*nr_bytes:])

def _tlv_object_list_to_dict(tlv_object_list):
    tags_to_ignore = ['6F', '70', '77', 'A5']
    tlv_object_dict = {}

    for (T, L, V) in tlv_object_list:
        if T not in tags_to_ignore:
            tlv_object_dict[T] = V
        if isinstance(V, list):
            tlv_object_dict.update(_tlv_object_list_to_dict(V))
    
    return tlv_object_dict



"""data.py
"""

# Standard library imports

# Third party imports

# Local application imports
from common import ber as _ber


tlv = {
    '42' : 'Issuer Identification Number',
    '4F' : 'Application Identifier',
    '50' : 'Application Label',
    '57' : 'Track 2 Equivalent Data',
    '5A' : 'Application Primary Account Number',
    '5F20' : 'Cardholder name',
    '5F24' : 'Application Expiration Date',
    '5F25' : 'Application Effective Date',
    '5F28' : 'Issuer country code',
    '5F2A' : 'Transaction Currency Code',
    '5F2D' : 'Language Preference',
    '5F30' : 'Service code',
    '5F34' : 'Application PAN Sequence Number',
    '5F36' : 'Transaction Currency Exponent',
    '5F50' : 'Issuer URL',
    '5F53' : 'International Bank Account Number',
    '5F54' : 'Bank Identifier Code (BIC)',
    '5F55' : 'Issuer Country Code (alpha2 format)',
    '5F56' : 'Issuer Country Code (alpha3 format)',
    '61' : 'Application Template',
    '6F' : 'File Control Information Template',
    '70' : 'Application Template',
    '71' : 'Issuer Script Template 1',
    '72' : 'Issuer script template 2',
    '73' : 'Directory Discretionary Template',
    '77' : 'Response Message Template Format 2',
    '80' : 'Response Message Template Format 1',
    '81' : 'Amount, Authorised (Binary)',
    '82' : 'Application interchange profile',
    '83' : 'Command Template',
    '84' : 'Dedicated File (DF) Name',
    '86' : 'Issuer Script Command',
    '87' : 'Application Priority Indicator',
    '88' : 'Short File Identifier',
    '89' : 'Authorisation code',
    '8A' : 'Authorisation response code',
    '8C' : 'Card Risk Management Data Object List 1',
    '8D' : 'Card Risk Management Data Object List 2',
    '8E' : 'Cardholder Verification Method (CVM) List',
    '8F' : 'Certification authority public key index',
    '90' : 'Issuer Public Key Certificate',
    '91' : 'Issuer authentication data',
    '92' : 'Issuer Public Key Remainder',
    '93' : 'Signed Static Application Data',
    '94' : 'Application File Locator',
    '95' : 'Terminal verification results',
    '97' : 'Transaction Certificate Data Object List',
    '98' : 'Transaction Certificate (TC) Hash Value',
    '99' : 'Transaction PIN Data',
    '9A' : 'Transaction Date',
    '9B' : 'Transaction Status Information',
    '9C' : 'Transaction Type',
    '9D' : 'Directory Definition File (DDF) Name',
    '9F01' : 'Acquirer Identifier',
    '9F02' : 'Amount, Authorised (Numeric)',
    '9F03' : 'Amount, Other (Numeric)',
    '9F04' : 'Amount, Other (Binary)',
    '9F05' : 'Application Discretionary Data',
    '9F06' : 'Application Identifier (AID) - Terminal',
    '9F07' : 'Application Usage Control',
    '9F08' : 'Application Version Number',
    '9F09' : 'Terminal application version number',
    '9F0B' : 'Cardholder Name Extended',
    '9F0D' : 'Issuer Action Code - Default',
    '9F0E' : 'Issuer Action Code - Denial',
    '9F0F' : 'Issuer Action Code - Online',
    '9F10' : 'Issuer Application Data',
    '9F11' : 'Issuer Code Table Index',
    '9F12' : 'Application Preferred Name',
    '9F13' : 'Last Online Application Transaction Counter (ATC) Register',
    '9F14' : 'Lower Consecutive Offline Limit',
    '9F15' : 'Merchant Category Code',
    '9F16' : 'Merchant Identifier',
    '9F17' : 'PIN Try Counter',
    '9F18' : 'Issuer Script Identifier',
    '9F1A' : 'Terminal Country Code',
    '9F1B' : 'Terminal Floor Limit',
    '9F1C' : 'Terminal Identification',
    '9F1D' : 'Terminal Risk Management Data',
    '9F1E' : 'Interface Device (IFD) Serial Number',
    '9F1F' : 'Track 1 Discretionary Data',
    '9F20' : 'Track 2 Discretionary Data',
    '9F21' : 'Transaction Time',
    '9F22' : 'Certification Authority Public Key Index',
    '9F23' : 'Upper Consecutive Offline Limit',
    '9F26' : 'Application Cryptogram',
    '9F27' : 'Cryptogram Information Data',
    '9F2D' : 'ICC PIN Encipherment Public Key Certificate',
    '9F2E' : 'ICC PIN Encipherment Public Key Exponent',
    '9F2F' : 'ICC PIN Encipherment Public Key Remainder',
    '9F32' : 'Issuer Public Key Exponent',
    '9F33' : 'Terminal Capabilities',
    '9F34' : 'Cardholder Verification Method (CVM) Results',
    '9F35' : 'Terminal Type',
    '9F36' : 'Application Transaction Counter',
    '9F37' : 'Unpredictable Number',
    '9F38' : 'Processing Options Data Object List',
    '9F39' : 'Point-of-Service (POS) Entry Mode',
    '9F3A' : 'Amount, Reference Currency',
    '9F3B' : 'Application Reference Currency',
    '9F3C' : 'Transaction Reference Currency Code',
    '9F3D' : 'Transaction Reference Currency Exponent',
    '9F40' : 'Additional Terminal Capabilities',
    '9F41' : 'Transaction Sequence Counter',
    '9F42' : 'Application Currency Code',
    '9F43' : 'Application Reference Currency Exponent',
    '9F44' : 'Application Currency Exponent',
    '9F45' : 'Data Authentication Code',
    '9F46' : 'ICC Public Key Certificate',
    '9F47' : 'ICC Public Key Exponent',
    '9F48' : 'ICC Public Key Remainder',
    '9F49' : 'Dynamic Data Authentication Data Object List',
    '9F4A' : 'Static Data Authentication Tag List',
    '9F4B' : 'Signed Dynamic Application Data',
    '9F4C' : 'ICC Dynamic Number',
    '9F4D' : 'Log Entry',
    '9F4E' : 'Merchant Name and Location',
    '9F4F' : 'Log Format',
    '9F53' : 'Transaction category code',
    'A5' : 'File Control Information (FCI) Proprietary Template',
    'B1' : 'Transaction log information',
    'B2' : 'Transaction log records',
    'B3' : 'Log totals',
    'BE' : 'Connection data 2',
    'BF0C' : 'File Control Information (FCI) Issuer Discretionary Data',
    'Issuer Identification Number' : '42',
    'Application Identifier' : '4F',
    'Application Label' : '50',
    'Track 2 Equivalent Data' : '57',
    'Application Primary Account Number' : '5A',
    'Cardholder name' : '5F20',
    'Application Expiration Date' : '5F24',
    'Application Effective Date' : '5F25',
    'Issuer country code' : '5F28',
    'Transaction Currency Code' : '5F2A',
    'Language Preference' : '5F2D',
    'Service code' : '5F30',
    'Application PAN Sequence Number' : '5F34',
    'Transaction Currency Exponent' : '5F36',
    'Issuer URL' : '5F50',
    'International Bank Account Number' : '5F53',
    'Bank Identifier Code (BIC)' : '5F54',
    'Issuer Country Code (alpha2 format)' : '5F55',
    'Issuer Country Code (alpha3 format)' : '5F56',
    'Application Template' : '61',
    'File Control Information Template' : '6F',
    'Application Template' : '70',
    'Issuer Script Template 1' : '71',
    'Issuer script template 2' : '72',
    'Directory Discretionary Template' : '73',
    'Response Message Template Format 2' : '77',
    'Response Message Template Format 1' : '80',
    'Amount, Authorised (Binary)' : '81',
    'Application interchange profile' : '82',
    'Command Template' : '83',
    'Dedicated File (DF) Name' : '84',
    'Issuer Script Command' : '86',
    'Application Priority Indicator' : '87',
    'Short File Identifier' : '88',
    'Authorisation code' : '89',
    'Authorisation response code' : '8A',
    'Card Risk Management Data Object List 1' : '8C',
    'Card Risk Management Data Object List 2' : '8D',
    'Cardholder Verification Method (CVM) List' : '8E',
    'Certification authority public key index' : '8F',
    'Issuer Public Key Certificate' : '90',
    'Issuer authentication data' : '91',
    'Issuer Public Key Remainder' : '92',
    'Signed Static Application Data' : '93',
    'Application File Locator' : '94',
    'Terminal verification results' : '95',
    'Transaction Certificate Data Object List' : '97',
    'Transaction Certificate (TC) Hash Value' : '98',
    'Transaction PIN Data' : '99',
    'Transaction Date' : '9A',
    'Transaction Status Information' : '9B',
    'Transaction Type' : '9C',
    'Directory Definition File (DDF) Name' : '9D',
    'Acquirer Identifier' : '9F01',
    'Amount, Authorised (Numeric)' : '9F02',
    'Amount, Other (Numeric)' : '9F03',
    'Amount, Other (Binary)' : '9F04',
    'Application Discretionary Data' : '9F05',
    'Application Identifier (AID) - Terminal' : '9F06',
    'Application Usage Control' : '9F07',
    'Application Version Number' : '9F08',
    'Terminal application version number' : '9F09',
    'Cardholder Name Extended' : '9F0B',
    'Issuer Action Code - Default' : '9F0D',
    'Issuer Action Code - Denial' : '9F0E',
    'Issuer Action Code - Online' : '9F0F',
    'Issuer Application Data' : '9F10',
    'Issuer Code Table Index' : '9F11',
    'Application Preferred Name' : '9F12',
    'Last Online Application Transaction Counter (ATC) Register' : '9F13',
    'Lower Consecutive Offline Limit' : '9F14',
    'Merchant Category Code' : '9F15',
    'Merchant Identifier' : '9F16',
    'PIN Try Counter' : '9F17',
    'Issuer Script Identifier' : '9F18',
    'Terminal Country Code' : '9F1A',
    'Terminal Floor Limit' : '9F1B',
    'Terminal Identification' : '9F1C',
    'Terminal Risk Management Data' : '9F1D',
    'Interface Device (IFD) Serial Number' : '9F1E',
    'Track 1 Discretionary Data' : '9F1F',
    'Track 2 Discretionary Data' : '9F20',
    'Transaction Time' : '9F21',
    'Certification Authority Public Key Index' : '9F22',
    'Upper Consecutive Offline Limit' : '9F23',
    'Application Cryptogram' : '9F26',
    'Cryptogram Information Data' : '9F27',
    'ICC PIN Encipherment Public Key Certificate' : '9F2D',
    'ICC PIN Encipherment Public Key Exponent' : '9F2E',
    'ICC PIN Encipherment Public Key Remainder' : '9F2F',
    'Issuer Public Key Exponent' : '9F32',
    'Terminal Capabilities' : '9F33',
    'Cardholder Verification Method (CVM) Results' : '9F34',
    'Terminal Type' : '9F35',
    'Application Transaction Counter' : '9F36',
    'Unpredictable Number' : '9F37',
    'Processing Options Data Object List' : '9F38',
    'Point-of-Service (POS) Entry Mode' : '9F39',
    'Amount, Reference Currency' : '9F3A',
    'Application Reference Currency' : '9F3B',
    'Transaction Reference Currency Code' : '9F3C',
    'Transaction Reference Currency Exponent' : '9F3D',
    'Additional Terminal Capabilities' : '9F40',
    'Transaction Sequence Counter' : '9F41',
    'Application Currency Code' : '9F42',
    'Application Reference Currency Exponent' : '9F43',
    'Application Currency Exponent' : '9F44',
    'Data Authentication Code' : '9F45',
    'ICC Public Key Certificate' : '9F46',
    'ICC Public Key Exponent' : '9F47',
    'ICC Public Key Remainder' : '9F48',
    'Dynamic Data Authentication Data Object List' : '9F49',
    'Static Data Authentication Tag List' : '9F4A',
    'Signed Dynamic Application Data' : '9F4B',
    'ICC Dynamic Number' : '9F4C',
    'Log Entry' : '9F4D',
    'Merchant Name and Location' : '9F4E',
    'Log Format' : '9F4F',
    'Transaction category code' : '9F53',
    'File Control Information (FCI) Proprietary Template' : 'A5',
    'Transaction log information' : 'B1',
    'Transaction log records' : 'B2',
    'Log totals' : 'B3',
    'Connection data 2' : 'BE',
    'File Control Information (FCI) Issuer Discretionary Data' : 'BF0C',
    'AC' : '9F26',
    'AFL' : '94',
    'AID' : '4F',
    'AIP' : '82',
    'ATC' : '9F36',
    'AUC' : '9F07',
    'CDOL1' : '8C',
    'CDOL2' : '8D',
    'CID' : '9F27',
    'DDOL' : '9F49',
    'IBAN' : '5F53',
    'IIN' : '42',
    'LATC' : '9F13',
    'LCOL' : '9F14',
    'MCC' : '9F15',
    'PAN' : '5A',
    'PDOL' : '9F38',
    'PSN' : '5F34',
    'PTC' : '9F17',
    'SFI' : '88',
    'TDOL' : '97',
    'Track2' : '57',
    'UCOL' : '9F23',
    'UN' : '9F37'
}

def AFL(tlv_objects_list):
    """AFL():
    """
    return tlv_objects_list.get('94', None)

def ATC(tlv_objects_list):
    """ATC():
    """
    return tlv_objects_list.get('9F36', None)

def PDOL(tlv_objects_list):
    """PDOL():
    """
    return tlv_objects_list.get('9F38', None)

def PSN(tlv_objects_list):
    """PSN():
    """
    return tlv_objects_list.get('5F34', None)

def Track2(tlv_objects_list):
    """Track2():
    """
    return tlv_objects_list.get('57', None)

def get_object(object_name: str, tlv_objects_list):
    """get_object():
    """
    return tlv_objects_list.get(tlv[object_name], None)

def PDOL_data(pdol: str, tlv_objects_list):
    """PDOL_data():
    """
    pdol , _ = _ber.parse_dol(pdol)
    return ''.join(map(lambda o: tlv_objects_list[o[0]], pdol))


#
# Helper functions (assume a clean 'hstr' as input)
#




"""bitstr.py

Generic functions working on strings containing only '0', '1', and '_' characters
"""

# Standard library imports
import math as _math

# Third party imports

# Local application imports


def clean(bstr_in: str, command_name='()', bstr_name='') -> str:
    """clean():

    - Removes '_' characters
    - Raises exceptions on illegal characters
    """
    bstr_out = ''
    for bit in bstr_in:
        if bit in '01':
            bstr_out += bit
        elif bit in ' _':
            pass
        else:
            if bstr_name == '':
                raise TypeError(F"{command_name}: Unknown bit: {bit}")
            else:
                raise TypeError(
                    F"{command_name}: Unknown bit in {command_name}: {bit}")

    return bstr_out


def to_hstr(bstr: str, command_name='()', bstr_name='') -> str:
    """to_hstr():
    """
    bstr = clean(bstr, command_name, bstr_name)
    length = _math.ceil(len(bstr)/4)

    return F"{int(bstr, 2):X}".rjust(length, '0')


def xor(bstr_a: str, bstr_b: str, command_name='()', bstr_name='') -> str:
    """xor(bstr_a, bstr_b):
    """
    bstr_a = clean(bstr_a, command_name, bstr_name)
    bstr_b = clean(bstr_b, command_name, bstr_name)
    length = max(len(bstr_a), len(bstr_b))

    return F"{int(bstr_a, 2)^int(bstr_b, 2):b}".rjust(length, "0")



"""modes.py: Block cipher modes of operation
"""

# Standard library imports

# Third party imports

# Local application imports
import common.hstr as _hstr

def cbc_encryption(enc, key, block, iv):
    """cbc_encryption()
    """
    tmp = iv
    ciphertext = ''
    for i in range(len(block)//16):
        plaintext = block[i*16:(i+1)*16]
        tmp = enc(key, _hstr.xor(tmp, plaintext))
        ciphertext += tmp

    return ciphertext


def cbc_decryption(dec, key, block, iv):
    """cbc_decryption()
    """
    tmp = iv
    plaintext = ''
    for i in range(len(block)//16):
        ciphertext = block[i*16:(i+1)*16]
        plaintext += _hstr.xor(tmp, dec(key, ciphertext))
        tmp = ciphertext

    return plaintext



"""hstr.py

Generic functions working on strings containing only
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    'A', 'B', 'C', 'D', 'E', 'F', 'a', 'b', 'c', 'd', 'e', 'f',
    and '_' characters
"""

# Standard library imports

# Third party imports

# Local application imports
from common.str import clean as _clean

def clean(hstr_in: str, command_name='()', hstr_name='') -> str:
    """clean(hstr)

    - Capitalizes hexadecimal digits
    - Removes '_' characters
    - Raises exceptions on illegal characters
    """
    return _clean(hstr_in, '0123456789ABCDEFabcdef', '_', command_name, hstr_name).upper()


def to_bitstr(hstr: str, command_name='()', hstr_name='') -> str:
    """to_bitstr(hstr): converts 'hex strings' to 'bit strings'
    """
    bstr = ''

    for nibble in clean(hstr, command_name, hstr_name):
        bstr += F"{int(nibble, 16):04b}"

    return bstr


def xor(hstr_a: str, hstr_b: str, command_name='()', hstr_name='') -> str:
    """xor(hstr_a, hstr_b):
    """
    hstr_a = clean(hstr_a, command_name, hstr_name)
    hstr_b = clean(hstr_b, command_name, hstr_name)
    length = max(len(hstr_a), len(hstr_b))
    hstr_a = hstr_a.rjust(length, '0')
    hstr_b = hstr_b.rjust(length, '0')
    xor_hstr = ''

    for (nibble_a, nibble_b) in zip(hstr_a, hstr_b):
        xor_hstr += F"{int(nibble_a, 16)^int(nibble_b, 16):X}"

    return xor_hstr

def and_(hstr_a: str, hstr_b: str, command_name='()', hstr_name='') -> str:
    """and_(hstr_a, hstr_b):
    """
    hstr_a = clean(hstr_a, command_name, hstr_name)
    hstr_b = clean(hstr_b, command_name, hstr_name)
    length = max(len(hstr_a), len(hstr_b))
    hstr_a = hstr_a.rjust(length, '0')
    hstr_b = hstr_b.rjust(length, '0')
    and_hstr = ''

    for (nibble_a, nibble_b) in zip(hstr_a, hstr_b):
        and_hstr += F"{int(nibble_a, 16)&int(nibble_b, 16):X}"

    return and_hstr

def not_hstr(hstr_in: str, command_name='()', hstr_name='') -> str:
    """not(hstr):
    """
    hstr_in = clean(hstr_in, command_name, hstr_name)
    hstr_out = ''

    for nibble in hstr_in:
        hstr_out += xor(nibble, 'F')

    return hstr_out


def to_dec(hstr: str, command_name='()', hstr_name='') -> int:
    """to_dec():
    """
    return int(clean(hstr, command_name, hstr_name), 16)


def to_intlist(hstr: str, command_name='()', hstr_name='') -> list[int]:
    """to_intlist():
    """
    hstr = clean(hstr, command_name, hstr_name)
    if (len(hstr) % 2) != 0:
        raise ValueError(
            F"{command_name}: uneven length for {hstr_name}: {len(hstr)}")

    bytelist = []
    for i in range(len(hstr)//2):
        bytelist.append(int(hstr[i*2:(i+1)*2], 16))

    return bytelist


def minimum(hstr_a: str, hstr_b: str) -> str:
    """minimum():
    """
    hstr_a = clean(hstr_a)
    hstr_b = clean(hstr_b)

    if int(hstr_a, 16) == min(int(hstr_a, 16), int(hstr_b, 16)):
        return hstr_a

    return hstr_b

def split_by_length(hstr:str, length:int) -> list[str]:
    """split_by(): 
    """
    if len(hstr) % length != 0:
        # Error if the string is not a multiple of the length divider
        return []
    
    nr_splits = len(hstr)//length

    return [hstr[i*length:(i+1)*length] for i in range(nr_splits)]

def dscan_decimalize(hstr: str) -> str:
    """dscan_decimalize(): double scan decimalization
    """
    dstr1 = ''
    dstr2 = ''
    for h in clean(hstr):
        if h.isdigit():
            print(F"{h} is a digit")
            dstr1 = dstr1 + h
        else:
            print(F"{h} is not a digit")
            dstr2 = dstr2 + F"{int(h, 16)-10}"

    return dstr1 + dstr2
